import logging
import time
from datetime import datetime, timedelta, timezone
from typing import List, Optional, Tuple, Union

from pydantic import BaseModel, PrivateAttr  # pylint: disable=no-name-in-module
from sqlalchemy import Column, DateTime, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm.session import Session
from sxm.models import XMChannel, XMLiveChannel

COOLDOWN_SHORT = 10
COOLDOWN_MED = 60
COOLDOWN_LONG = 600

Base = declarative_base()


class DBSong(Base):
    __tablename__ = "songs"

    guid = Column(String, primary_key=True)
    title = Column(String, index=True)
    artist = Column(String, index=True)
    album = Column(String, nullable=True)
    air_time = Column(DateTime)
    channel = Column(String)
    file_path = Column(String)
    image_url = Column(String, nullable=True)


class Song(BaseModel):
    guid: str
    title: str
    artist: str
    album: Optional[str]
    air_time: datetime
    channel: str
    file_path: str
    image_url: Optional[str]

    class Config:
        orm_mode = True

    @property
    def air_time_smart(self):
        return self.air_time.replace(tzinfo=timezone.utc)

    @staticmethod
    def get_pretty_name(
        title: Optional[str], artist: Optional[str], bold: bool = False
    ) -> str:
        """Returns a formatted name of song"""

        if title is None:
            title = ""

        if artist is None:
            artist = ""

        mod = ""
        if bold:
            mod = "**"

        return f'{mod}"{title}"{mod} by {mod}{artist}{mod}'

    @property
    def pretty_name(self) -> str:
        """Returns a formatted name of song"""

        return Song.get_pretty_name(self.title, self.artist)

    @property
    def bold_name(self) -> str:
        """Returns a formatted name of song"""

        return Song.get_pretty_name(self.title, self.artist, True)


class DBEpisode(Base):
    __tablename__ = "episodes"

    guid = Column(String, primary_key=True)
    title = Column(String, index=True)
    show = Column(String, nullable=True, index=True)
    air_time = Column(DateTime)
    channel = Column(String)
    file_path = Column(String)
    image_url = Column(String, nullable=True)


class Episode(BaseModel):
    guid: str
    title: str
    show: str
    air_time: datetime
    channel: str
    file_path: str
    image_url: Optional[str]

    class Config:
        orm_mode = True

    @staticmethod
    def get_pretty_name(
        title: Optional[str],
        show: Optional[str],
        air_time: Optional[datetime],
        bold: bool = False,
    ) -> str:
        """Returns a formatted name of show"""

        if title is None:
            title = ""

        if show is None:
            show = ""

        mod = ""
        if bold:
            mod = "**"

        if air_time is None:
            f'{mod}"{title}"{mod} ({show})'
        return f'{mod}"{title}"{mod} ({show}) from {air_time}'

    @property
    def pretty_name(self) -> str:
        """Returns a formatted name of show"""

        return Episode.get_pretty_name(self.title, self.show, self.air_time)

    @property
    def bold_name(self) -> str:
        """Returns a formatted name of show"""

        return Episode.get_pretty_name(self.title, self.show, self.air_time, True)


class PlayerState(BaseModel):
    stream_url: Optional[str] = None
    stream_channel: Optional[str] = None
    processed_folder: Optional[str] = None
    db_reset: bool = False
    sxm_running: bool = False
    player_name: Optional[str] = None

    _db: Optional[Session] = PrivateAttr(None)
    _raw_channels: Optional[List[dict]] = PrivateAttr(None)
    _raw_live: Optional[dict] = PrivateAttr(None)
    _channels: Optional[List[XMChannel]] = PrivateAttr(None)
    _live: Optional[XMLiveChannel] = PrivateAttr(None)
    _failures: int = PrivateAttr(0)
    _cooldown: float = PrivateAttr(0)
    _last_failure: float = PrivateAttr(0)
    _start_time: Optional[datetime] = PrivateAttr(None)
    _time_offset: Optional[timedelta] = PrivateAttr(None)

    @property
    def stream_data(self) -> Tuple[Optional[str], Optional[str]]:
        return (self.stream_channel, self.stream_url)

    def update_stream_data(self, value: Tuple[Optional[str], Optional[str]]) -> None:
        self.stream_channel = value[0]
        self.stream_url = value[1]

    @property
    def channels(self) -> List[XMChannel]:
        """Returns list of `XMChannel`"""

        if self._channels is None:
            if self._raw_channels is None:
                return []
            self._channels = []
            for channel in self._raw_channels:
                self._channels.append(XMChannel.from_dict(channel))
        return self._channels

    def update_channels(self, value: Optional[List[dict]]) -> None:
        """
        Sets channel key in internal `_raw_channels`.
        """

        self._channels = None
        self._raw_channels = value

        if self._raw_channels is None:
            self.stream_url = None
            self.stream_channel = None

    def get_raw_channels(self) -> Optional[List[dict]]:
        return self._raw_channels

    @property
    def db(self) -> Union[Session, None]:
        if self._db is None and self.processed_folder is not None:
            from .utils import init_db

            self._db = init_db(self.processed_folder, reset=self.db_reset)
        return self._db

    @property
    def live(self) -> Union[XMLiveChannel, None]:
        """Returns current `XMLiveChannel`"""

        return self._live

    def update_live(self, value: dict) -> None:
        """Sets live key in internal `_raw_live`."""

        now = datetime.now(timezone.utc)
        self._live = None
        self._raw_live = value

        if self._raw_live is not None:
            self._live = XMLiveChannel.from_dict(self._raw_live)

            if self._live.tune_time is not None:
                self._time_offset = now - self._live.tune_time

            if self._start_time is None:
                if self._live.tune_time is None:
                    self._start_time = now
                else:
                    self._start_time = self._live.tune_time
        else:
            self._time_offset = 0
            self._start_time = None

    def get_raw_live(
        self,
    ) -> Tuple[Optional[datetime], Optional[timedelta], Optional[dict]]:
        return (self._start_time, self._time_offset, self._raw_live)

    def set_raw_live(
        self,
        live_data: Tuple[Optional[datetime], Optional[timedelta], Optional[dict]],
    ):
        self._start_time = live_data[0]
        self._time_offset = live_data[1]
        self._raw_live = live_data[2]

        if self._raw_live is not None:
            self._live = XMLiveChannel.from_dict(self._raw_live)

    @property
    def radio_time(self) -> Union[datetime, None]:
        """Returns current time for the radio"""

        if self.live is None:
            return None

        now = datetime.now(timezone.utc)
        if self._time_offset is not None:
            return now - self._time_offset
        return now

    @property
    def start_time(self) -> Optional[datetime]:
        """Returns the start time for the current SiriusXM channel"""

        if self.live is None:
            return None
        return self._start_time

    @property
    def is_connected(self) -> bool:
        is_connected = self._raw_channels is not None
        if is_connected and time.monotonic() - self._last_failure > 300:
            self._failures = 0
        return is_connected

    @property
    def can_connect(self) -> bool:
        return time.monotonic() > self._cooldown

    def mark_attempt(self, logger: logging.Logger) -> float:
        if self.can_connect:
            self.mark_failure()
            extra_seconds = self.increase_cooldown()

            logger.info(
                "Attempting to connect SXM Client (next in " f"{extra_seconds} seconds)"
            )
            return True
        return False

    def increase_cooldown(self) -> float:
        extra_seconds = 0
        if self._failures < 3:
            extra_seconds = COOLDOWN_SHORT
        elif self._failures < 5:
            extra_seconds = COOLDOWN_MED
        else:
            extra_seconds = COOLDOWN_LONG

        self._cooldown = time.monotonic() + extra_seconds

        return extra_seconds

    def mark_failure(self) -> float:
        self._failures += 1
        self._last_failure = time.monotonic()
        return self._cooldown

    def get_channel(self, name: str) -> Union[XMChannel, None]:
        """Returns channel from list of `channels` with given name"""

        name = name.lower()
        for channel in self.channels:
            if (
                channel.name.lower() == name
                or channel.id.lower() == name
                or channel.channel_number == name
            ):
                return channel
        return None



from typing import List, Optional, Tuple, Type

import click

from sxm_player.models import PlayerState
from sxm_player.players.base import BasePlayer, Option
from sxm_player.runner import Runner
from sxm_player.workers import BaseWorker, CLIPlayerWorker


class CLIPlayer(BasePlayer):
    params: List[click.Parameter] = [
        Option(
            "--channel-id",
            required=True,
            type=str,
            help="SXM Channel to Player",
        ),
        Option(
            "--filename",
            default="player.mp3",
            type=click.Path(dir_okay=False, writable=True, resolve_path=True),
            help="Path for output mp3 file",
        ),
    ]

    @staticmethod
    def get_params() -> List[click.Parameter]:
        return CLIPlayer.params

    @staticmethod
    def get_worker_args(
        runner: Runner, state: PlayerState, **kwargs
    ) -> Optional[Tuple[Type[BaseWorker], str, dict]]:

        context = click.get_current_context()
        params = {
            "filename": context.meta["filename"],
            "stream_protocol": "udp",
            "sxm_status": state.sxm_running,
            "stream_data": (context.meta["channel_id"], state.stream_url),
            "channels": state.get_raw_channels(),
            "raw_live_data": state.get_raw_live(),
        }

        return (CLIPlayerWorker, "cli_player", params)



import typer

from sxm_player.cli import main
from sxm_player.command import ConfigCommandClass


def start():
    app = typer.Typer()
    app.command(
        cls=ConfigCommandClass,
        context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
    )(main)
    app()


if __name__ == "__main__":
    start()



import httpx

from ..queue import Event, EventMessage
from .base import SXMLoopedWorker

__all__ = ["StatusWorker"]

CHECK_INTERVAL = 30


class StatusWorker(SXMLoopedWorker):
    NAME = "status_check"

    _ip: str
    _port: int
    _delay: float = 30.0
    _failures: int = 0

    def __init__(self, port: int, ip: str, *args, **kwargs):

        super().__init__(*args, **kwargs)

        if ip == "0.0.0.0":  # nosec
            ip = "127.0.0.1"

        self._ip = ip
        self._port = port

    def loop(self):
        self.check_sxm()

    def check_sxm(self):
        if self._state.sxm_running:
            self._log.debug("Checking SXM Client")
            r = httpx.get(f"http://{self._ip}:{self._port}/channels/")

            if r.is_error:
                # adjust delay to check more often
                self._delay = 5.0
                self._failures += 1
                if self._failures > 3:
                    self.push_event(
                        EventMessage(self.name, Event.RESET_SXM, "bad status check")
                    )
            else:
                self._delay = 30.0
                self._failures = 0
                self.push_event(
                    EventMessage(self.name, Event.UPDATE_CHANNELS, r.json())
                )



import logging
import time
from datetime import datetime, timedelta
from multiprocessing import synchronize
from typing import List, Optional, Tuple

from ..models import PlayerState
from ..queue import Event, EventMessage, Queue
from ..signals import default_signal_handler, init_signals, interupt_signal_handler

__all__ = [
    "BaseWorker",
    "InterruptableWorker",
    "LoopedWorker",
    "SXMStatusSubscriber",
    "HLSStatusSubscriber",
    "SXMLoopedWorker",
]


class BaseWorker:
    NAME = "worker"

    _log: logging.Logger

    name: str = NAME
    int_handler: staticmethod = staticmethod(default_signal_handler)
    term_handler: staticmethod = staticmethod(default_signal_handler)
    startup_event: synchronize.Event
    shutdown_event: synchronize.Event
    local_shutdown_event: synchronize.Event

    def __init__(
        self,
        startup_event: synchronize.Event,
        shutdown_event: synchronize.Event,
        local_shutdown_event: synchronize.Event,
        event_queue: Queue,
        name: str = "worker",
        *args,
        **kwargs,
    ):
        self._log = logging.getLogger(f"sxm_player.{name}")

        self.name = name

        self.startup_event = startup_event
        self.shutdown_event = shutdown_event
        self.local_shutdown_event = local_shutdown_event
        self.event_queue = event_queue

    def init_signals(self):
        self._log.debug("Entering init_signals")
        signal_object = init_signals(
            self.shutdown_event, self.int_handler, self.term_handler
        )
        return signal_object

    def start(self):
        self.init_signals()

        self.startup_event.set()
        return self.run()

    def run(self):
        raise NotImplementedError("run method not implemented")

    def push_event(self, event: EventMessage):
        success = self.event_queue.safe_put(event)

        if not success:
            self._log.error(f"Could not pass event: {event.msg_src}, {event.msg_type}")


class InterruptableWorker(BaseWorker):
    int_handler: staticmethod = staticmethod(interupt_signal_handler)
    term_handler: staticmethod = staticmethod(interupt_signal_handler)


class LoopedWorker(BaseWorker):
    _delay: float = 1

    def run(self):
        self.setup()

        while not self.shutdown_event.is_set():
            time.sleep(self._delay)
            self.loop()

        self.cleanup()

    def cleanup(self):
        pass

    def setup(self):
        pass

    def loop(self):
        raise NotImplementedError("loop method not implemented")


class SXMStatusSubscriber:
    sxm_status_queue: Queue

    def __init__(self, sxm_status_queue):
        self.sxm_status_queue = sxm_status_queue


class HLSStatusSubscriber:
    hls_stream_queue: Queue

    def __init__(self, hls_stream_queue):
        self.hls_stream_queue = hls_stream_queue


class EventedWorker(LoopedWorker):
    _last_loop: float = 0
    _event_queues: List[Queue]
    _event_delay: float = 0

    def run(self):
        self.setup()

        try:
            while (
                not self.shutdown_event.is_set()
                and not self.local_shutdown_event.is_set()
            ):
                for queue in self._event_queues:
                    event = queue.safe_get()

                    if event:
                        self._log.debug(
                            f"Received event: {event.msg_src}, "
                            f"{event.msg_type.name}"
                        )
                        self._handle_event(event)

                if time.monotonic() > (self._last_loop + self._delay):
                    self.loop()
                    self._last_loop = time.monotonic()
        except Exception as e:
            self._log.error(f"Exception occurred in {self.name}: {e}")

        self.cleanup()

    def _handle_event(self, event: EventMessage):
        raise NotImplementedError("_handle_event method not implemented")


class SXMLoopedWorker(EventedWorker, SXMStatusSubscriber):
    _state: PlayerState

    def __init__(self, sxm_status: bool, *args, **kwargs):
        sxm_status_queue = kwargs.pop("sxm_status_queue")
        SXMStatusSubscriber.__init__(self, sxm_status_queue)
        super().__init__(*args, **kwargs)

        self._state = PlayerState()
        self._state.sxm_running = sxm_status
        self._event_queues = [self.sxm_status_queue]

    def _handle_event(self, event: EventMessage):
        if event.msg_type == Event.SXM_STATUS:
            self._state.sxm_running = event.msg
        else:
            self._log.warning(
                f"Unknown event received: {event.msg_src}, {event.msg_type}"
            )


class HLSLoopedWorker(EventedWorker, HLSStatusSubscriber):
    _state: PlayerState

    def __init__(
        self,
        stream_data: Tuple[Optional[str], Optional[str]] = (None, None),
        channels: Optional[List[dict]] = None,
        raw_live_data: Tuple[
            Optional[datetime], Optional[timedelta], Optional[dict]
        ] = (
            None,
            None,
            None,
        ),
        *args,
        **kwargs,
    ):
        hls_stream_queue = kwargs.pop("hls_stream_queue")
        HLSStatusSubscriber.__init__(self, hls_stream_queue)
        super().__init__(*args, **kwargs)

        self._event_queues = [self.hls_stream_queue]

        self._state = PlayerState()
        self._state.update_stream_data(stream_data)
        self._state.update_channels(channels)
        self._state.set_raw_live(raw_live_data)

    def _handle_event(self, event: EventMessage):
        if event.msg_type == Event.HLS_STREAM_STARTED:
            self._state.update_stream_data(event.msg)
        elif event.msg_type == Event.UPDATE_METADATA:
            self._state.set_raw_live(event.msg)
        elif event.msg_type == Event.UPDATE_CHANNELS:
            self._state.update_channels(event.msg)
        elif event.msg_type == Event.KILL_HLS_STREAM:
            self.local_shutdown_event.set()
        else:
            self._log.warning(
                f"Unknown event received: {event.msg_src}, {event.msg_type}"
            )


class ComboLoopedWorker(EventedWorker, SXMStatusSubscriber, HLSStatusSubscriber):
    _state: PlayerState

    def __init__(
        self,
        sxm_status: bool,
        stream_data: Tuple[Optional[str], Optional[str]],
        raw_live_data: Tuple[Optional[datetime], Optional[timedelta], Optional[dict]],
        *args,
        **kwargs,
    ):
        sxm_status_queue = kwargs.pop("sxm_status_queue")
        SXMStatusSubscriber.__init__(self, sxm_status_queue)

        hls_stream_queue = kwargs.pop("hls_stream_queue")
        HLSStatusSubscriber.__init__(self, hls_stream_queue)
        super().__init__(*args, **kwargs)

        self._event_queues = [self.hls_stream_queue, self.sxm_status_queue]

        self._state = PlayerState()
        self._state.sxm_running = sxm_status
        self._state.update_stream_data(stream_data)
        self._state.set_raw_live(raw_live_data)



import inspect
from typing import List, Optional, Tuple, Type

import click

from sxm_player.models import PlayerState
from sxm_player.runner import Runner
from sxm_player.workers import BaseWorker


class BasePlayer:
    @staticmethod
    def get_params() -> List[click.Parameter]:
        return []

    @staticmethod
    def get_worker_args(
        runner: Runner, state: PlayerState, **kwargs
    ) -> Optional[Tuple[Type[BaseWorker], str, dict]]:
        return None


class Option(click.Option):
    def __init__(self, *params_decls, **kwargs):
        if "help" in kwargs:
            kwargs["help"] = inspect.cleandoc(kwargs["help"])

        super().__init__(params_decls, **kwargs)



import os
from datetime import timedelta
from time import monotonic
from typing import Dict, List, Optional, Union

from sxm.models import XMCutMarker, XMEpisodeMarker, XMSong

from sxm_player.models import DBEpisode, DBSong
from sxm_player.utils import (
    from_fs_datetime,
    get_art_thumb_url,
    get_art_url_by_size,
    get_files,
    splice_file,
)
from sxm_player.workers.archiver import ARCHIVE_CHUNK
from sxm_player.workers.base import HLSLoopedWorker

__all__ = ["ProcessorWorker"]

MAX_DUPLICATE_COUNT = 3
CUT_PADDING = timedelta(seconds=20)


class ProcessorWorker(HLSLoopedWorker):
    """Runs song/show processor"""

    NAME = "processor"

    _delay: float = ARCHIVE_CHUNK.total_seconds()

    def __init__(
        self,
        processed_folder: str,
        archive_folder: str,
        reset_songs: bool,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.processed_folder = processed_folder
        self.archive_folder = archive_folder

        # run in 90 seconds and run ~30 seconds after Archiver
        self._last_loop = monotonic() + 90 - ARCHIVE_CHUNK.total_seconds()
        self._state.processed_folder = self.processed_folder
        self._state.db_reset = reset_songs

        # force db initialization
        self._state.db

        os.makedirs(self.processed_folder, exist_ok=True)
        os.makedirs(self.archive_folder, exist_ok=True)

    def loop(self) -> None:
        self._delay = ARCHIVE_CHUNK.total_seconds()

        if (
            self._state.stream_channel is None
            or self._state.live is None
            or self.archive_folder is None
        ):
            return None

        channel_archive = os.path.join(self.archive_folder, self._state.stream_channel)
        os.makedirs(channel_archive, exist_ok=True)

        archives = {}
        archive_files = get_files(channel_archive)
        for archive_file in archive_files:
            file_parts = archive_file.split(".")
            archive_key = f"{file_parts[1]}.{file_parts[2]}"
            archives[archive_key] = os.path.join(channel_archive, archive_file)
        self._log.debug(f"found {len(archives.keys())}")

        processed_songs = self._process_cuts(archives, self._state.live.song_cuts)
        processed_shows = self._process_cuts(archives, self._state.live.episode_markers)

        self._log.info(f"processed: {processed_songs} songs, {processed_shows} shows")

    def _path_filter(self, word: str) -> str:
        """Filters out known words to call issues for creating
        names for folders/files"""

        return (
            word.replace("Counterfeit.", "Counterfeit")
            .replace("F**ker", "Fucker")
            .replace("Trust?", "Trust")
            .replace("P.O.D.", "POD")
            .replace("//", "-")
            .replace("@", "")
            .replace("(", "")
            .replace(")", "")
            .strip()
        )

    def _process_cut(
        self, archives: Dict[str, str], cut: Union[XMCutMarker, XMEpisodeMarker]
    ) -> bool:
        """Processes `archives` to splice out an
        instance of `XMMarker` if it exists"""

        if (
            self.processed_folder is None
            or self._state.stream_channel is None
            or self._state.db is None
        ):
            return False

        archive = None
        start = cut.time + CUT_PADDING
        splice_start = timedelta(seconds=0)
        padded_duration = cut.duration + CUT_PADDING
        end = start + padded_duration
        splice_end = timedelta(seconds=0)

        for archive_key, archive_file in archives.items():
            archive_start, archive_end = [
                from_fs_datetime(i) for i in archive_key.split(".")
            ]

            if archive_start < start and archive_end > end:
                archive = archive_file
                splice_start = start - archive_start
                splice_end = splice_start + padded_duration
                break

        if archive is not None:
            self._log.debug(f"found archive {archive}")

            title = None
            album_or_show = None
            album_url = None
            artist = None
            filename = None
            folder = os.path.join(self.processed_folder, self._state.stream_channel)

            if isinstance(cut, XMEpisodeMarker):
                title = self._path_filter(
                    cut.episode.long_title or cut.episode.medium_title
                )

                if cut.episode.show is not None:
                    album_or_show = self._path_filter(
                        cut.episode.show.long_title or cut.episode.show.medium_title
                    )
                    album_url = get_art_thumb_url(cut.episode.show.arts)

                filename = (
                    f'{title}.{cut.time.strftime("%Y-%m-%d-%H.%M")}' f".{cut.guid}.mp3"
                )
                folder = os.path.join(folder, "shows")
            elif isinstance(cut.cut, XMSong):
                title = self._path_filter(cut.cut.title)
                artist = self._path_filter(cut.cut.artists[0].name)

                if cut.cut.album is not None and cut.cut.album.title is not None:
                    album_or_show = self._path_filter(cut.cut.album.title)
                    album_url = get_art_url_by_size(cut.cut.album.arts, "MEDIUM")

                filename = f"{title}.{cut.guid}.mp3"
                folder = os.path.join(folder, "songs", artist)
            else:
                return False

            if album_or_show is not None:
                folder = os.path.join(folder, album_or_show)

            os.makedirs(folder, exist_ok=True)
            path: Optional[str] = os.path.join(folder, filename)
            self._log.debug(f"{cut.duration}: {path}")
            if path is not None:
                self._log.debug(
                    f"Splice song: (Song: {start}, {end}, {cut.duration}), "
                    f"(Archive: {archive}, {splice_start}, {splice_end}"
                )
                path = splice_file(
                    archive,
                    path,
                    int(splice_start.total_seconds()),
                    int(splice_end.total_seconds()),
                )

            if path is not None:
                if os.path.getsize(path) < 1000:
                    self._log.error(
                        f"spliced file too small, deleting {path}: {archive}"
                    )
                    os.remove(path)
                    return False

                is_song = False
                if isinstance(cut, XMEpisodeMarker):
                    db_item: Union[DBSong, DBEpisode] = DBEpisode(
                        guid=cut.guid,
                        title=title,
                        show=album_or_show,
                        air_time=cut.time,
                        channel=self._state.stream_channel,
                        file_path=path,
                        image_url=album_url,
                    )
                elif isinstance(cut.cut, XMSong):
                    is_song = True
                    db_item = DBSong(
                        guid=cut.guid,
                        title=title,
                        artist=artist,
                        album=album_or_show,
                        air_time=cut.time,
                        channel=self._state.stream_channel,
                        file_path=path,
                        image_url=album_url,
                    )
                else:
                    return False

                self._state.db.add(db_item)
                self._state.db.commit()
                self._log.debug(f"inserted cut {is_song}: {db_item.guid}")
                return True
        return False

    def _process_cuts(
        self,
        archives: Dict[str, str],
        cuts: Union[List[XMCutMarker], List[XMEpisodeMarker]],
    ) -> int:
        """Processes `archives` to splice out any
        instance of `XMMarker` if it exists"""

        if self._state.live is None or self._state.db is None:
            return 0

        processed = 0
        for cut in cuts:
            if cut.duration == 0.0:
                continue

            db_item: Union[DBSong, DBEpisode, None] = None
            if isinstance(cut, XMEpisodeMarker):
                db_item = (
                    self._state.db.query(DBEpisode).filter_by(guid=cut.guid).first()
                )
            elif isinstance(cut.cut, XMSong):
                existing = (
                    self._state.db.query(DBSong)
                    .filter_by(title=cut.cut.title, artist=cut.cut.artists[0].name)
                    .all()
                )

                if len(existing) >= MAX_DUPLICATE_COUNT:
                    continue

                db_item = self._state.db.query(DBSong).filter_by(guid=cut.guid).first()

            if db_item is not None:
                continue

            title: Optional[str] = None
            if isinstance(cut, XMEpisodeMarker):
                title = cut.episode.long_title or cut.episode.medium_title
            elif isinstance(cut.cut, XMSong):
                title = cut.cut.title

            if title is None:
                title = "unknown"
            self._log.debug(
                f"processing {title}: " f"{cut.time}: {cut.duration}" f"{cut.guid}"
            )
            success = self._process_cut(archives, cut)

            if success:
                processed += 1
        return processed



import logging
import time
from multiprocessing import Event, Process, synchronize
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Type, TypeVar

from sxm_player.queue import Queue
from sxm_player.signals import default_signal_handler, init_signals
from sxm_player.utils import configure_root_logger
from sxm_player.workers import BaseWorker, HLSStatusSubscriber, SXMStatusSubscriber

STOP_WAIT_SECS = 3.0
STARTUP_WAIT_SECS = 10.0


def _sleep_secs(max_sleep, end_time=999_999_999_999_999.9):
    # Calculate time left to sleep, no less than 0
    return max(0.0, min(end_time - time.monotonic(), max_sleep))


def worker_wrapper(
    worker_class: Type[BaseWorker],
    log_level: str,
    log_file: Optional[Path],
    startup_event: synchronize.Event,
    shutdown_event: synchronize.Event,
    local_shutdown_event: synchronize.Event,
    event_queue: Queue,
    sxm_status_queue: Optional[Queue],
    hls_stream_queue: Optional[Queue],
    name: str,
    *args,
    **kwargs,
):

    kwargs["name"] = name
    kwargs["startup_event"] = startup_event
    kwargs["shutdown_event"] = shutdown_event
    kwargs["local_shutdown_event"] = local_shutdown_event
    kwargs["event_queue"] = event_queue

    if issubclass(worker_class, SXMStatusSubscriber):
        kwargs["sxm_status_queue"] = sxm_status_queue

    if issubclass(worker_class, HLSStatusSubscriber):
        kwargs["hls_stream_queue"] = hls_stream_queue

    configure_root_logger(log_level, log_file)

    worker = worker_class(*args, **kwargs)
    return worker.start()


class Worker:
    startup_event: synchronize.Event
    shutdown_event: synchronize.Event
    local_shutdown_event: synchronize.Event
    process: Process
    name: str
    sxm_status_queue: Optional[Queue] = None
    hls_stream_queue: Optional[Queue] = None

    def __init__(
        self,
        logger: logging.Logger,
        log_level: str,
        log_file: Optional[Path],
        worker_class: Type[BaseWorker],
        shutdown_event: synchronize.Event,
        event_queue: Queue,
        sxm_status_queue: Optional[Queue],
        hls_stream_queue: Optional[Queue],
        name: str,
        debug: bool,
        *args,
        **kwargs,
    ):

        self.name = name
        self.log = logger
        self.startup_event = Event()
        self.shutdown_event = shutdown_event
        self.local_shutdown_event = Event()
        self.sxm_status_queue = sxm_status_queue
        self.hls_stream_queue = hls_stream_queue

        self.process = Process(
            target=worker_wrapper,
            args=(
                worker_class,
                log_level,
                log_file,
                self.startup_event,
                self.shutdown_event,
                self.local_shutdown_event,
                event_queue,
                sxm_status_queue,
                hls_stream_queue,
                name,
                *args,
            ),
            kwargs=kwargs,
        )

        self.log.debug(f"Starting worker: {name}")
        self.process.start()

        timeout = STARTUP_WAIT_SECS
        if debug:
            timeout = 999_999_999_999_999

        started = self.startup_event.wait(timeout=timeout)

        self.log.debug(f"Startup Event: {name} got {started}")
        if not started:
            self.terminate()
            raise RuntimeError(
                f"Process {name} failed to startup after {timeout} seconds"
            )

    def full_stop(self, wait_time=STOP_WAIT_SECS):
        self.log.debug(f"stopping: {self.name}")
        self.local_shutdown_event.set()
        self.process.join(wait_time)
        if self.process.is_alive():
            self.terminate()

    def terminate(self):
        self.log.debug(f"Terminating: {self.name}")

        NUM_TRIES = 3
        tries = NUM_TRIES
        while tries and self.process.is_alive():
            self.process.terminate()
            time.sleep(0.01)
            tries -= 1

        if self.process.is_alive():
            self.log.error(
                f"Failed to terminate {self.name} after {NUM_TRIES} attempts"
            )
            return False
        else:
            self.log.info(
                f"Terminated {self.name} after {NUM_TRIES - tries} attempt(s)"
            )
            return True


RunnerType = TypeVar("RunnerType", bound="Runner")


class Runner:
    workers: Dict[str, Worker]
    queues: List[Queue]
    shutdown_event: synchronize.Event
    event_queue: Queue
    log: logging.Logger
    log_level: str
    log_file: Optional[Path]

    def __init__(self, log_file: Optional[Path], debug: bool):
        self.workers = {}
        self.queues = []
        self.shutdown_event = Event()
        self.event_queue = self.create_queue()

        log_level = "INFO"
        if debug:
            log_level = "DEBUG"

        configure_root_logger(log_level, log_file)
        self.log = logging.getLogger("sxm_player")
        self.log_level = log_level
        self.log_file = log_file

    def __enter__(self: RunnerType) -> RunnerType:
        init_signals(
            self.shutdown_event, default_signal_handler, default_signal_handler
        )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.log.error(
                f"Exception: {exc_val}", exc_info=(exc_type, exc_val, exc_tb)
            )

        self.stop_workers()
        self.stop_queues()

        # -- Don't eat exceptions that reach here.
        return not exc_type

    def stop_workers(self) -> Tuple[int, int]:
        self.shutdown_event.set()
        end_time = time.monotonic() + STOP_WAIT_SECS
        num_terminated = 0
        num_failed = 0

        # Gracefully let the process try to stop
        for worker in self.workers.values():
            join_secs = _sleep_secs(STOP_WAIT_SECS, end_time)
            worker.process.join(join_secs)

        still_running: Dict[str, Worker] = {}
        while len(self.workers.keys()) > 0:
            first_key = list(self.workers.keys())[0]
            worker = self.workers.pop(first_key)
            terminated, failed, running = self.stop_worker(worker)

            num_terminated += terminated
            num_failed += failed
            if running:
                still_running[worker.name] = worker

        self.workers = still_running
        return num_failed, num_terminated

    def stop_worker(self, worker) -> Tuple[int, int, bool]:
        terminated = 0
        failed = 0
        running = False

        if worker.process.is_alive():
            if worker.terminate():
                terminated = 1
            else:
                running = True
        else:
            exitcode = worker.process.exitcode
            if exitcode:
                self.log.error(
                    (f"Process {worker.name} ended with " f"exitcode {exitcode}")
                )
                terminated = 2
            else:
                self.log.debug(f"Process {worker.name} stopped successfully")

        return (terminated, failed, running)

    def stop_queues(self) -> int:
        num_items_left = 0
        # -- Clear the queues list and close all associated queues
        for q in self.queues:
            num_items_left += sum(1 for __ in q.drain())
            q.close()

        # -- Wait for all queue threads to stop
        while self.queues:
            q = self.queues.pop(0)
            q.join_thread()
        return num_items_left

    def create_queue(self, *args, **kwargs) -> Queue:
        queue = Queue(*args, **kwargs)
        self.queues.append(queue)

        return queue

    def create_worker(
        self, worker_class: Type[BaseWorker], name: str, *args, **kwargs
    ) -> Worker:

        sxm_status_queue: Optional[Queue] = None
        hls_stream_queue: Optional[Queue] = None
        if issubclass(worker_class, SXMStatusSubscriber):
            sxm_status_queue = self.create_queue()

        if issubclass(worker_class, HLSStatusSubscriber):
            hls_stream_queue = self.create_queue()

        worker = Worker(
            self.log,
            self.log_level,
            self.log_file,
            worker_class,
            self.shutdown_event,
            self.event_queue,
            sxm_status_queue,
            hls_stream_queue,
            name,
            self.log_level == "DEBUG",
            *args,
            **kwargs,
        )
        self.workers[name] = worker
        return worker



from sxm_player.workers.cli import CLIPlayerWorker
from sxm_player.workers.hls import HLSWorker
from sxm_player.workers.processor import ProcessorWorker
from sxm_player.workers.server import ServerWorker
from sxm_player.workers.status import StatusWorker

from .archiver import ArchiveWorker
from .base import (
    BaseWorker,
    ComboLoopedWorker,
    EventedWorker,
    HLSLoopedWorker,
    HLSStatusSubscriber,
    InterruptableWorker,
    LoopedWorker,
    SXMLoopedWorker,
    SXMStatusSubscriber,
)

# debug.py is not included in published package
try:
    from sxm_player.debug.worker import DebugWorker
except ImportError:
    DebugWorker = None  # type: ignore

__all__ = [
    "ArchiveWorker",
    "BaseWorker",
    "CLIPlayerWorker",
    "ComboLoopedWorker",
    "DebugWorker",
    "EventedWorker",
    "HLSLoopedWorker",
    "HLSStatusSubscriber",
    "HLSWorker",
    "InterruptableWorker",
    "LoopedWorker",
    "ProcessorWorker",
    "ServerWorker",
    "StatusWorker",
    "SXMLoopedWorker",
    "SXMStatusSubscriber",
]



from typing import Optional, Tuple, Type

from sxm_player.models import PlayerState
from sxm_player.players.base import BasePlayer
from sxm_player.runner import Runner
from sxm_player.workers import BaseWorker, DebugWorker


class DebugPlayer(BasePlayer):
    @staticmethod
    def get_worker_args(
        runner: Runner, state: PlayerState, **kwargs
    ) -> Optional[Tuple[Type[BaseWorker], str, dict]]:
        return (DebugWorker, "debug", {})



import os
from datetime import datetime, timedelta, timezone
from time import monotonic
from typing import Dict, Optional, Tuple, Union

from sxm_player.queue import Event, EventMessage
from sxm_player.utils import create_fs_datetime, get_files, splice_file
from sxm_player.workers.base import HLSLoopedWorker

__all__ = ["ArchiveWorker"]

ARCHIVE_DROPOFF = timedelta(hours=24)
ARCHIVE_CHUNK = timedelta(minutes=10)
ARCHIVE_BUFFER = timedelta(seconds=5)


class ArchiveWorker(HLSLoopedWorker):
    NAME = "archiver"

    stream_folder: str
    archive_folder: str
    last_size: Dict[str, int] = {}

    _delay: float = ARCHIVE_CHUNK.total_seconds()

    def __init__(self, stream_folder: str, archive_folder: str, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # run in 11 minutes (ARCHIVE_CHUNK + 1 minute) to ensure a
        # full 10 minutes of audio exists
        self._last_loop = monotonic() + 60

        self.stream_folder = stream_folder
        self.archive_folder = archive_folder

        os.makedirs(self.stream_folder, exist_ok=True)
        os.makedirs(self.archive_folder, exist_ok=True)

    def loop(self) -> None:
        if self._state.stream_channel is None or self.stream_folder is None:
            self.local_shutdown_event.set()
            return

        deleted = 0
        archived = None
        stream_files = get_files(self.stream_folder)

        for stream_file in stream_files:
            abs_path = os.path.join(self.stream_folder, stream_file)
            archived, removed = self._process_file(abs_path)
            deleted += removed

        self._log.info(
            f"archived: deleted files: {deleted}, " f"archived file: {archived}"
        )

    def _process_file(self, abs_path) -> Tuple[Optional[str], int]:
        archived = None
        deleted = 0

        if not self._validate_name(abs_path):
            deleted += 1
        elif self._validate_size(abs_path):
            archived, removed = self._process_stream_file(abs_path)
            deleted += removed

        return (archived, deleted)

    def _validate_name(self, stream_file) -> bool:
        stream_file = os.path.basename(stream_file)
        file_parts = stream_file.split(".")
        if file_parts[-1] != "mp3" or file_parts[0] != self._state.stream_channel:
            os.remove(stream_file)
            return False
        return True

    def _validate_size(self, stream_file) -> bool:
        if not self._check_size(stream_file):
            self._log.error("archive not increasing, resetting channel")
            self.push_event(EventMessage(self.name, Event.KILL_HLS_STREAM, None))
            return False
        return True

    def _check_size(self, abs_path: str) -> bool:
        current = os.path.getsize(abs_path)
        if (
            self.last_size.get(abs_path) is not None
            and self.last_size[abs_path] == current
        ):
            return False

        self.last_size[abs_path] = current
        return True

    def _delete_old_archives(
        self, archive_folder: str, archive_base: str, current_file: str
    ) -> int:
        """Deletes any old versions of archive that is about to be made"""

        archive_files = get_files(archive_folder)

        now = datetime.now(timezone.utc)
        removed: int = 0
        for archive_file in archive_files:
            abs_path = os.path.join(archive_folder, archive_file)
            access_time = datetime.fromtimestamp(os.path.getatime(abs_path)).replace(
                tzinfo=timezone.utc
            )

            age = now - access_time
            if (
                archive_file.startswith(archive_base) and archive_file != current_file
            ) or age > ARCHIVE_DROPOFF:

                self._log.debug(f"deleted old archive: {abs_path}")
                os.remove(abs_path)
                removed += 1
        return removed

    def _process_stream_file(self, abs_path: str) -> Tuple[Union[str, None], int]:
        """Processes stream file by creating an archive from
        it if necessary"""

        channel_id = self._state.stream_channel
        if channel_id is None or self.archive_folder is None:
            return (None, 0)
        channel_archive = os.path.join(self.archive_folder, channel_id)

        now = self._state.radio_time or datetime.now(timezone.utc)
        start_time = self._state.start_time or datetime.fromtimestamp(
            os.path.getatime(abs_path)
        ).replace(tzinfo=timezone.utc)

        max_archive_cutoff = now - ARCHIVE_BUFFER
        creation_time = start_time + ARCHIVE_BUFFER

        time_elapsed = max_archive_cutoff - creation_time
        archive_chunks = int(time_elapsed / ARCHIVE_CHUNK)
        if archive_chunks > 0:
            os.makedirs(channel_archive, exist_ok=True)
            time_elapsed = archive_chunks * ARCHIVE_CHUNK
            archive_cutoff = creation_time + time_elapsed

            archive_base = f"{channel_id}.{create_fs_datetime(creation_time)}"
            archive_filename = (
                f"{archive_base}.{create_fs_datetime(archive_cutoff)}.mp3"
            )
            archive_output = os.path.join(channel_archive, archive_filename)
            if os.path.exists(archive_output):
                return (None, 0)

            removed = self._delete_old_archives(
                channel_archive, archive_base, archive_filename
            )
            return (
                splice_file(
                    abs_path,
                    archive_output,
                    int(ARCHIVE_BUFFER.total_seconds()),
                    int((ARCHIVE_BUFFER + time_elapsed).total_seconds()),
                ),
                removed,
            )
        return (None, 0)



import logging
from typing import Callable

from aiohttp import web
from sxm import QualitySize, RegionChoice, SXMClient, make_http_handler

from ..queue import Event, EventMessage
from ..signals import TerminateInterrupt
from .base import InterruptableWorker

__all__ = ["ServerWorker"]


class ServerWorker(InterruptableWorker):
    """SXM Client proxy server for sxm-player to interface with"""

    NAME = "sxm"

    _ip: str
    _port: int
    sxm: SXMClient

    def __init__(
        self,
        port: int,
        ip: str,
        username: str,
        password: str,
        region: RegionChoice,
        quality: QualitySize,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self._port = port
        self._ip = ip

        self.sxm = SXMClient(
            username=username,
            password=password,
            region=region,
            quality=quality,
            update_handler=self._make_update_handler(),
        )

        self.sxm.authenticate()
        self.sxm.configuration

    def _make_update_handler(self) -> Callable[[dict], None]:
        """Returns update handler to be called by
        `SXMClient.get_playlist` when a HLS playlist updates"""

        def update_handler(data: dict) -> None:
            self.push_event(EventMessage(self.name, Event.UPDATE_METADATA, data))

        return update_handler

    def send_channel_list(self):
        channels = self.sxm.get_channels()

        self.push_event(EventMessage(self.name, Event.UPDATE_CHANNELS, channels))

    def run(self) -> None:
        """Runs SXM proxy server"""

        self.send_channel_list()

        request_logger = logging.getLogger("sxm_player.server.request")
        request_logger._info = request_logger.info  # type: ignore
        request_logger.info = request_logger.debug  # type: ignore

        app = web.Application()
        app.router.add_get("/{_:.*}", make_http_handler(self.sxm.async_client))
        try:
            self._log.info(f"{self.name} has started on http://{self._ip}:{self._port}")
            web.run_app(
                app,
                host=self._ip,
                port=self._port,
                access_log=request_logger,
                print=None,  # type: ignore
            )
        except (KeyboardInterrupt, TerminateInterrupt):
            pass



import os
import tempfile
import time
from typing import Optional, Tuple

from ..queue import Event, EventMessage
from ..utils import FFmpeg
from .base import SXMLoopedWorker

__all__ = ["HLSWorker"]


FFMPEG_COMMAND = "ffmpeg -loglevel warning -f hls -i {} " "-f mpegts {} "
FFMPEG_PROTOCOLS = [
    "udp",
    # "rtsp",
    # "rtmp",
    "unix",
]


class HLSWorker(SXMLoopedWorker, FFmpeg):
    NAME = "hls"

    channel_id: str
    stream_file: Optional[str] = None
    stream_url: str
    stream_protocol: str
    playback_url: str

    _start: float = 0

    def __init__(
        self,
        ip: str,
        port: int,
        channel_id: str,
        stream_folder: Optional[str],
        stream_protocol: str = "udp",
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.stream_url = f"http://{ip}:{port}/{channel_id}.m3u8"
        self.channel_id = channel_id

        port = port + 1

        self.playback_url, output_options = self._get_playback_url(
            stream_protocol, ip, port, channel_id
        )

        log_message = f"playing {self.stream_url}"
        if stream_folder is not None:
            self.stream_file = os.path.join(stream_folder, f"{channel_id}.mp3")

            if os.path.exists(self.stream_file):
                os.remove(self.stream_file)

            log_message += f" ({self.stream_file})"

            output_options = f"{output_options} file:/{self.stream_file}"

        self._log.info(log_message)
        self.command = FFMPEG_COMMAND.format(self.stream_url, output_options)
        self.start_ffmpeg()

    def _get_playback_url(
        self, stream_protocol: str, ip: str, port: int, channel_id: str
    ) -> Tuple[str, str]:
        if stream_protocol not in FFMPEG_PROTOCOLS:
            self._log.warning(
                f"Unknown stream protocol: {stream_protocol}, " "defaulting to udp"
            )
            stream_protocol = "udp"

        output_options = ""
        if stream_protocol == "udp":
            playback_url = f"udp://{ip}:{port}"
            output_options = f"{playback_url}"
        # elif stream_protocol == "rtsp":
        #     playback_url = f"rtsp://127.0.0.1:{port}"
        #     output_options = f"-rtsp_flags listen {playback_url}"
        # elif stream_protocol == "rtmp":
        #     playback_url = f"rtmp://127.0.0.1:{port}"
        #     output_options = f"-listen 1 {playback_url}"
        else:
            socket_file = os.path.join(tempfile.gettempdir(), f"{channel_id}.sock")
            if os.path.exists(socket_file):
                os.remove(socket_file)

            playback_url = f"unix:/{socket_file}"
            output_options = f"-listen 1 {playback_url}"

        return (playback_url, output_options)

    def setup(self):
        self._start = time.monotonic()

        self.push_event(
            EventMessage(
                self.name,
                Event.HLS_STREAM_STARTED,
                (self.channel_id, self.playback_url),
            )
        )

    def loop(self):
        now = time.monotonic()

        if not self._state.sxm_running:
            self._log.info(f"SXM Client is dead, stopping {self.name}")
            self.local_shutdown_event.set()
            return
        elif not self.check_process():
            self._log.info(f"ffmpeg process is not active, stopping {self.name}")
            self.local_shutdown_event.set()
            return
        elif (
            now - self._start > 5
            and self.stream_file is not None
            and not os.path.exists(self.stream_file)
        ):
            self._log.info(f"stream file missing, stopping {self.name}")
            self.local_shutdown_event.set()
            return

        lines = self.read_errors()

        if len(lines) > 0:
            self._log.debug(f"adding {len(lines)} of stderr to shared memory")
            self.push_event(EventMessage(self.name, Event.HLS_STDERROR_LINES, lines))

    def cleanup(self):
        self.stop_ffmpeg()

        self.push_event(EventMessage(self.name, Event.KILL_HLS_STREAM, None))



import logging
import os
import select
import shlex
import subprocess  # nosec
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Union

import coloredlogs  # type: ignore
import psutil
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session
from sxm.models import XMArt, XMImage

from sxm_player.models import DBEpisode, DBSong

ACTIVE_PROCESS_STATUSES = [
    psutil.STATUS_RUNNING,
    psutil.STATUS_SLEEPING,
    psutil.STATUS_DISK_SLEEP,
]
FS_DATETIME_FORMAT = "%Y%m%d-%H%M%S%z"


unrelated_loggers = [
    "discord.client",
    "discord.gateway",
    "discord.http",
    "urllib3.connectionpool",
    "websockets.protocol",
]

logger = logging.getLogger("sxm_player.utils")


def init_db(
    base_folder: str,
    cleanup: Optional[bool] = True,
    reset: Optional[bool] = False,
) -> Session:
    """Initializes song database connection"""

    from .models import Base

    os.makedirs(base_folder, exist_ok=True)

    song_db = os.path.join(base_folder, "songs.db")

    if reset and os.path.exists(song_db):
        logger.info("Reseting database...")
        os.remove(song_db)

    db_engine = create_engine(f"sqlite:///{song_db}")
    Base.metadata.create_all(db_engine)
    db_session = sessionmaker(bind=db_engine)()

    if cleanup:
        removed = 0
        for song in db_session.query(DBSong).all():
            if not os.path.exists(song.file_path):
                removed += 1
                db_session.delete(song)

        for show in db_session.query(DBEpisode).all():
            if not os.path.exists(show.file_path):
                removed += 1
                db_session.delete(show)

        if removed > 0:
            logger.warn(f"deleted missing songs/shows: {removed}")
            db_session.commit()

    logger.info("Database initalized")
    return db_session


def get_art_url_by_size(arts: List[XMArt], size: str) -> Optional[str]:
    for art in arts:
        if isinstance(art, XMImage) and art.size is not None and art.size == size:
            return art.url
    return None


def get_art_thumb_url(arts: List[XMArt]) -> Optional[str]:
    thumb: Optional[str] = None

    for art in arts:
        if (
            isinstance(art, XMImage)
            and art.height is not None
            and art.height > 100
            and art.height < 200
            and art.height == art.width
        ):
            # logo on dark is what we really want
            if art.name == "show logo on dark":
                thumb = art.url
                break
            # but it is not always there, so fallback image
            elif art.name == "image":
                thumb = art.url

    return thumb


def get_files(folder: str) -> List[str]:
    """Gets list of files in a folder"""

    dir_list = os.listdir(folder)

    files = []
    for dir_item in dir_list:
        abs_path = os.path.join(folder, dir_item)
        if os.path.isfile(abs_path):
            files.append(dir_item)

    return files


def splice_file(
    input_file: str, output_file: str, start_time: int, end_time: int
) -> Union[str, None]:
    """Splices a chunk off of the input file and saves it"""

    ffmpeg_command = 'ffmpeg -y -i "{}" -acodec copy -ss {} -to {} -loglevel fatal "{}"'
    args = shlex.split(
        ffmpeg_command.format(input_file, start_time, end_time, output_file)
    )

    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    try:
        subprocess.run(args, check=True)  # nosec
    except subprocess.CalledProcessError as e:
        logger.error(f"failed to split file: {e}")
        return None
    else:
        logger.info(f"spliced file: {output_file}")
        return output_file


def create_fs_datetime(dt):
    return dt.strftime(FS_DATETIME_FORMAT)


def from_fs_datetime(dt_string):
    return datetime.strptime(dt_string, FS_DATETIME_FORMAT)


def configure_root_logger(level: str, log_file: Optional[Path] = None):
    root_logger = logging.getLogger()
    if len(root_logger.handlers) == 0:
        if log_file is not None:
            fh = logging.FileHandler(log_file)
            formatter = logging.Formatter(
                "%(asctime)s %(name)s[%(process)d] %(levelname)s %(message)s"
            )
            fh.setLevel(level)
            fh.setFormatter(formatter)
            root_logger.addHandler(fh)
        coloredlogs.install(level=level, logger=root_logger)

    for logger in unrelated_loggers:
        logging.getLogger(logger).setLevel(logging.INFO)


class FFmpeg:
    command: str
    process: Optional[subprocess.Popen] = None

    _stderr_poll: Optional[select.poll] = None

    def start_ffmpeg(self) -> None:
        ffmpeg_args = shlex.split(self.command)

        self.process = subprocess.Popen(ffmpeg_args, stderr=subprocess.PIPE)  # nosec

        self._stderr_poll = select.poll()

        if self.process.stderr is not None:
            self._stderr_poll.register(self.process.stderr, select.POLLIN)

    def check_process(self) -> bool:
        if self.process is None:
            return False

        process = psutil.Process(self.process.pid)
        status = process.status()

        return status in ACTIVE_PROCESS_STATUSES

    def stop_ffmpeg(self) -> None:
        if self.process is None:
            return

        self.process.kill()
        if self.process.poll() is None:
            self.process.communicate()

        self.process = None

    def read_errors(self) -> List[str]:
        if self.process is None or self._stderr_poll is None:
            return []

        lines: List[str] = []
        while self._stderr_poll.poll(0.1):
            if self.process.stderr is not None:
                lines.append(self.process.stderr.readline().decode("utf8"))

        return lines



import os
from bdb import BdbQuit

from sxm_player.queue import Event, EventMessage
from sxm_player.workers.base import InterruptableWorker

__all__ = ["DebugWorker"]


class DebugWorker(InterruptableWorker):
    """DebugWorker to be used for development"""

    _num: int = 0

    def run(self) -> None:
        try:
            self.debug()
        except BdbQuit:
            self._log.error("No debugger to break for")

    def debug(self):
        self.nothing_to_see_here()

    def nothing_to_see_here(self):
        breakpoint()

    def play_channel(self, channel_id: str, protocol: str = "udp"):
        self._num += 1

        player_name = f"debug_player_{channel_id}{self._num}"
        filename = os.path.abspath(f"{player_name}.mp3")
        self.push_event(
            EventMessage(
                self.name,
                Event.DEBUG_START_PLAYER,
                (player_name, channel_id, filename, protocol),
            )
        )

    def stop_player(self, player_name, kill_hls=True):
        self.push_event(EventMessage(self.name, Event.DEBUG_STOP_PLAYER, player_name))

        if kill_hls:
            self.kill_hls()

    def trigger_hls(self, channel_id, protocol="udp"):
        self.push_event(
            EventMessage(self.name, Event.TRIGGER_HLS_STREAM, (channel_id, protocol))
        )

    def kill_hls(self):
        self.push_event(EventMessage(self.name, Event.KILL_HLS_STREAM, None))



from .base import BasePlayer, Option
from .cli import CLIPlayer

# debug.py is not included in published package
try:
    from sxm_player.debug.player import DebugPlayer
except ImportError:
    DebugPlayer = None  # type: ignore

__all__ = ["BasePlayer", "CLIPlayer", "DebugPlayer", "Option"]



import functools
import signal


class TerminateInterrupt(BaseException):
    pass


class SignalObject:
    MAX_TERMINATE_CALLED = 3

    def __init__(self, shutdown_event):
        self.terminate_called = 0
        self.shutdown_event = shutdown_event


def interupt_signal_handler(
    signal_object, exception_class, signal_num, current_stack_frame
):
    raise exception_class()


def default_signal_handler(
    signal_object, exception_class, signal_num, current_stack_frame
):
    signal_object.terminate_called += 1
    signal_object.shutdown_event.set()
    if signal_object.terminate_called >= signal_object.MAX_TERMINATE_CALLED:
        raise exception_class()


def init_signal(signal_num, signal_object, exception_class, handler):
    handler = functools.partial(handler, signal_object, exception_class)
    signal.signal(signal_num, handler)
    signal.siginterrupt(signal_num, False)


def init_signals(shutdown_event, int_handler, term_handler):
    signal_object = SignalObject(shutdown_event)
    init_signal(signal.SIGINT, signal_object, KeyboardInterrupt, int_handler)
    init_signal(signal.SIGTERM, signal_object, TerminateInterrupt, term_handler)
    return signal_object



import importlib
import inspect
from typing import List, Optional, Type, Union

import typer
import yaml
from typer.core import TyperCommand

from sxm_player.players import BasePlayer

BASE_PLAYER = "sxm_player.players.BasePlayer "


class ConfigCommandClass(TyperCommand):
    def make_context(self, info_name, args, parent=None, **extra):
        config_file = self.get_config_file(args)
        if config_file is not None:
            config = self.load_config(config_file)
            extra["default_map"] = config

        response = super().make_context(info_name, args, parent, **extra)
        return response

    def parse_args(self, context, args):
        parser = self.make_parser(context)
        options, _, _ = parser.parse_args(args=args.copy())

        extra_args: List[str] = []
        if "player_class" in options:
            player_class = self.get_player_class(context, options)
            if player_class is not None:
                player_params = player_class.get_params()
                self.params = self.params + player_params
                extra_args = [p.name for p in player_params]

        super().parse_args(context, args)
        self.map_extra_args(context, extra_args)
        return

    def map_extra_args(self, context, extra_args: List[str]):
        for arg in extra_args:
            if arg in context.params:
                param = context.params.pop(arg)
                context.meta[arg] = param

    def get_player_class(self, context, options) -> Optional[Type[BasePlayer]]:
        player_param = None
        player_class = None

        for param in self.params:
            if param.human_readable_name == "PLAYER_CLASS":
                player_param = param
                break

        if player_param is not None:
            player_param.expose_value = False
            player_class, _ = player_param.handle_parse_result(context, options, [])
            player_param.expose_value = True
        return player_class

    def get_config_file(self, args: list) -> Optional[str]:
        config_file = None
        index = -1

        try:
            index = args.index("-c")
        except ValueError:
            try:
                index = args.index("--config-file")
            except ValueError:
                pass

        index += 1
        if index != 0 and len(args) > index:
            config_file = args[index]

        return config_file

    def load_config(self, config_file: str) -> dict:
        with open(config_file) as f:
            return yaml.safe_load(f)


def _get_module(class_name):
    module_path = "sxm_player.players"

    if "." in class_name:
        try:
            module_path, class_name = class_name.rsplit(".", 1)
        except ValueError:
            raise typer.BadParameter(f"{class_name} is not a Python path")

    return (module_path, class_name)


def validate_player(value: Union[str, Type[BasePlayer]]):
    if value is None or not isinstance(value, str) and issubclass(value, BasePlayer):
        return value

    module_path, class_name = _get_module(value)

    try:
        module = importlib.import_module(module_path)
    except ModuleNotFoundError:
        raise typer.BadParameter(f"{module_path} is not a Python module")

    try:
        klass = getattr(module, class_name)
    except AttributeError:
        raise typer.BadParameter(f"{class_name} does not inside in {module_path}")

    if not inspect.isclass(klass):
        raise typer.BadParameter(f"{class_name} is not a class")

    if not issubclass(klass, BasePlayer):
        raise typer.BadParameter(f"{class_name} does not inherit from {BASE_PLAYER}")

    return klass



# -*- coding: utf-8 -*-

"""Console script for sxm_player."""
import os
from multiprocessing import set_start_method
from pathlib import Path
from typing import Optional, Type

import psutil
import typer
from sxm import QualitySize, RegionChoice
from sxm.cli import (
    OPTION_HOST,
    OPTION_PASSWORD,
    OPTION_PORT,
    OPTION_QUALITY,
    OPTION_REGION,
    OPTION_USERNAME,
    OPTION_VERBOSE,
)

from sxm_player import handlers
from sxm_player.command import validate_player
from sxm_player.models import PlayerState
from sxm_player.players import BasePlayer
from sxm_player.queue import Event, EventMessage
from sxm_player.runner import Runner
from sxm_player.utils import ACTIVE_PROCESS_STATUSES
from sxm_player.workers import ServerWorker, StatusWorker

OPTION_CONFIG_FILE = typer.Option(
    None,
    "-c",
    "--config-file",
    exists=True,
    file_okay=True,
    dir_okay=False,
    readable=True,
    resolve_path=True,
    help="Config file to read vars from",
)
OPTION_LOG_FILE = typer.Option(
    None,
    "-l",
    "--log-file",
    exists=True,
    file_okay=True,
    resolve_path=True,
    dir_okay=False,
    readable=True,
    help="Output log file",
)
OPTION_OUTPUT_FOLDER = typer.Option(
    None,
    "-o",
    "--output-folder",
    file_okay=False,
    dir_okay=True,
    readable=True,
    writable=True,
    resolve_path=True,
    envvar="SXM_OUTPUT_FOLDER",
    help="output folder to save stream off to as it plays them",
)
OPTION_RESET_SONGS = typer.Option(
    False,
    "-R",
    "--reset-songs",
    help="Reset processed song database",
)
ARG_PLAYER_CLASS = typer.Argument(
    None, callback=validate_player, help="Optional Player Class to use"
)


def main(
    config_file: Optional[Path] = OPTION_CONFIG_FILE,
    log_file: Optional[Path] = OPTION_LOG_FILE,
    verbose: bool = OPTION_VERBOSE,
    username: str = OPTION_USERNAME,
    password: str = OPTION_PASSWORD,
    region: RegionChoice = OPTION_REGION,
    quality: QualitySize = OPTION_QUALITY,
    port: int = OPTION_PORT,
    host: str = OPTION_HOST,
    output_folder: Optional[Path] = OPTION_OUTPUT_FOLDER,
    reset_songs: bool = OPTION_RESET_SONGS,
    player_class: Optional[str] = ARG_PLAYER_CLASS,
):
    """Command line interface for sxm-player"""

    if verbose:
        set_start_method("spawn")

    os.system("/usr/bin/clear")  # nosec

    klass: Optional[Type[BasePlayer]] = None
    if player_class is not None:
        klass = player_class  # type: ignore

    with Runner(log_file, verbose) as runner:
        state = PlayerState()

        runner.create_worker(
            StatusWorker,
            StatusWorker.NAME,
            port=port,
            ip=host,
            sxm_status=state.sxm_running,
        )

        if klass is not None:
            worker_args = klass.get_worker_args(**locals())
            if worker_args is not None:
                state.player_name = worker_args[1]
                runner.create_worker(worker_args[0], worker_args[1], **(worker_args[2]))

        while not runner.shutdown_event.is_set():
            event_loop(**locals())

    return 0


def spawn_sxm_worker(
    runner: Runner,
    host: str,
    port: int,
    username: str,
    password: str,
    region: RegionChoice,
    quality: QualitySize,
    **kwargs,
):
    runner.create_worker(
        ServerWorker,
        ServerWorker.NAME,
        port=port,
        ip=host,
        username=username,
        password=password,
        region=region,
        quality=quality,
    )


def event_loop(runner: Runner, state: PlayerState, **kwargs):
    if not state.is_connected:
        if state.mark_attempt(runner.log):
            spawn_sxm_worker(runner, **kwargs)

    event = runner.event_queue.safe_get()

    if not event:
        return

    runner.log.debug(f"Received event: {event.msg_src}, {event.msg_type.name}")

    was_connected: Optional[bool] = None
    if event.msg_src == ServerWorker.NAME:
        was_connected = state.is_connected

    handle_event(event=event, runner=runner, state=state, **kwargs)

    if was_connected is False and state.is_connected:
        if not was_connected and state.is_connected:
            runner.log.info(
                f"SXM Client started. {len(state.channels)} channels available"
            )

            state.sxm_running = True
            handlers.sxm_status_event(runner, Event.SXM_STATUS, state.sxm_running)

    check_player(runner, state)


def handle_event(event: EventMessage, **kwargs):
    runner = kwargs["runner"]
    debug = kwargs["verbose"]
    event_name = event.msg_type.name.lower()
    is_debug_event = event_name.startswith("debug")
    handler_name = f"handle_{event_name}_event"

    if hasattr(handlers, handler_name) and (not is_debug_event or debug):
        getattr(handlers, handler_name)(event, **kwargs)
    else:
        runner.log.warning(f"Unknown event received: {event.msg_src}, {event.msg_type}")


def check_player(runner: Runner, state: PlayerState):
    if state.player_name is not None:
        player = runner.workers.get(state.player_name)
        running = True

        if player is None:
            running = False
        else:
            process = psutil.Process(player.process.pid)
            if process.status() not in ACTIVE_PROCESS_STATUSES:
                running = False

        if not running:
            runner.log.info("Player has stopped, shutting down")
            runner.shutdown_event.set()



import multiprocessing.queues as mpq
import time
from enum import Enum, auto
from multiprocessing import get_context
from queue import Empty, Full
from typing import Any, Optional

DEFAULT_POLLING_TIMEOUT = 0.02


class Event(Enum):
    RESET_SXM = auto()
    SXM_STATUS = auto()
    UPDATE_CHANNELS = auto()
    UPDATE_METADATA = auto()
    HLS_STREAM_STARTED = auto()
    HLS_STDERROR_LINES = auto()
    TRIGGER_HLS_STREAM = auto()
    KILL_HLS_STREAM = auto()
    DEBUG_START_PLAYER = auto()
    DEBUG_STOP_PLAYER = auto()


class EventMessage:
    id: float  # noqa: A003
    msg_src: str
    msg_relay: str
    msg_type: Event
    msg: Any

    def __init__(self, msg_src, msg_type, msg, msg_relay=None):
        self.id = time.monotonic()
        self.msg_src = msg_src
        self.msg_relay = msg_relay
        self.msg_type = msg_type
        self.msg = msg

    def __str__(self):
        return f"{self.msg_src} - {self.msg_type}: {self.msg}"


class Queue(mpq.Queue):
    # -- See StackOverflow Article :
    #   https://stackoverflow.com/questions/39496554/cannot-subclass-multiprocessing-queue-in-python-3-5
    #
    # -- tldr; mp.Queue is a _method_ that returns an mpq.Queue object.  That
    # object requires a context for proper operation, so this __init__ does
    # that work as well.
    def __init__(self, *args, **kwargs):
        ctx = get_context()
        super().__init__(*args, **kwargs, ctx=ctx)

    def safe_get(
        self, timeout: float = DEFAULT_POLLING_TIMEOUT
    ) -> Optional[EventMessage]:
        try:
            if timeout is None:
                return self.get(block=False)
            else:
                return self.get(block=True, timeout=timeout)
        except Empty:
            return None

    def safe_put(
        self, item: EventMessage, timeout: float = DEFAULT_POLLING_TIMEOUT
    ) -> bool:
        try:
            self.put(item, block=False, timeout=timeout)
            return True
        except Full:
            return False

    def drain(self):
        item = self.safe_get()
        while item:
            yield item
            item = self.safe_get()

    def safe_close(self) -> int:
        num_left = sum(1 for __ in self.drain())
        self.close()
        self.join_thread()
        return num_left



import time
from typing import Optional

from ..queue import Event, EventMessage
from ..utils import FFmpeg
from .base import ComboLoopedWorker

FFMPEG_COMMAND = "ffmpeg -y -loglevel fatal -f mpegts -i {} {}"


class CLIPlayerWorker(ComboLoopedWorker, FFmpeg):
    channel_id: Optional[str]
    stream_protocol: str

    _event_cooldown: float = 0

    def __init__(self, filename: str, *args, stream_protocol: str = "udp", **kwargs):
        super().__init__(*args, **kwargs)

        self.channel_id = self._state.stream_channel
        self.stream_protocol = stream_protocol
        self.filename = filename

        if self.channel_id is None:
            raise RuntimeError("No channel_id or stream_url provided")

    def loop(self):
        if self._state.sxm_running and self._state.stream_url is not None:
            self._valid_stream_loop()
        else:
            self._invalid_stream_loop()

    def _valid_stream_loop(self):
        if self.process is None:
            if self._state.stream_url is not None:
                self._log.info(f"Starting new HLS player: {self._state.stream_url}")
                self.command = FFMPEG_COMMAND.format(
                    self._state.stream_url, self.filename
                )

                time.sleep(3)
                self._log.info(f"CLI Player start: {self.name}")
                self.start_ffmpeg()
        elif not self.check_process():
            self._log.info("ffmpeg process is not active, removing ffmpeg process")
            self.cleanup()
        else:
            # read errors must be ran to prevent deadlock
            self.read_errors()

    def _invalid_stream_loop(self):
        if self.process is None:
            if self._state.sxm_running and self._state.stream_url is None:
                now = time.monotonic()
                if now > self._event_cooldown:
                    self._event_cooldown = now + 10
                    self._log.info(f"Starting new HLS stream: {self.channel_id}")
                    self.push_event(
                        EventMessage(
                            self.name,
                            Event.TRIGGER_HLS_STREAM,
                            (self.channel_id, self.stream_protocol),
                        )
                    )
        else:
            self._log.info("stream is dead, killing ffmpeg")
            self.cleanup()

    def cleanup(self):
        self.stop_ffmpeg()
        self._state.update_stream_data((None, None))

    def _handle_event(self, event: EventMessage):
        if event.msg_type == Event.SXM_STATUS:
            self._state.sxm_running = event.msg
        elif event.msg_type == Event.HLS_STREAM_STARTED:
            self._state.update_stream_data(event.msg)
        elif event.msg_type == Event.UPDATE_METADATA:
            self._state.set_raw_live(event.msg)
        elif event.msg_type == Event.UPDATE_CHANNELS:
            self._state.update_channels(event.msg)
        elif event.msg_type == Event.KILL_HLS_STREAM:
            self._log.info("stream is stopping, killing ffmpeg")
            self.cleanup()
        else:
            self._log.warning(
                f"Unknown event received: {event.msg_src}, {event.msg_type}"
            )



import os
from typing import Optional

from sxm_player.models import PlayerState
from sxm_player.queue import Event, EventMessage
from sxm_player.runner import Runner, Worker
from sxm_player.workers import (
    ArchiveWorker,
    CLIPlayerWorker,
    HLSWorker,
    ProcessorWorker,
    ServerWorker,
)


def hls_start_event(runner: Runner, stream_data: tuple, src: Optional[str] = None):
    hls_event(runner, Event.HLS_STREAM_STARTED, stream_data, src=src)


def hls_kill_event(runner: Runner, src: Optional[str] = None):
    hls_event(runner, Event.KILL_HLS_STREAM, None, src=src)


def hls_metadata_event(runner: Runner, live_data: tuple, src: Optional[str] = None):
    hls_event(runner, Event.UPDATE_METADATA, live_data, src=src)


def hls_channels_event(
    runner: Runner, channels: Optional[list], src: Optional[str] = None
):
    hls_event(runner, Event.UPDATE_CHANNELS, channels, src=src)


def hls_event(runner: Runner, event: Event, data, src: Optional[str] = None):
    for worker in runner.workers.values():
        if worker.hls_stream_queue is not None:
            if src is None:
                push_event(
                    runner,
                    worker,
                    "hls_stream_queue",
                    EventMessage("main", event, data),
                )
            else:
                push_event(
                    runner,
                    worker,
                    "hls_stream_queue",
                    EventMessage(src, event, data, msg_relay="main"),
                )


def sxm_status_event(
    runner: Runner, event: Event, status: bool, src: Optional[str] = None
):
    for worker in runner.workers.values():
        if worker.sxm_status_queue is not None:
            if src is None:
                push_event(
                    runner,
                    worker,
                    "sxm_status_queue",
                    EventMessage("main", event, status),
                )
            else:
                push_event(
                    runner,
                    worker,
                    "sxm_status_queue",
                    EventMessage(src, event, status, msg_relay="main"),
                )


def push_event(runner: Runner, worker: Worker, queue_name: str, event: EventMessage):

    success = getattr(worker, queue_name).safe_put(event)

    if not success:
        runner.log.error(f"Could not pass status event to {worker.name}")


def handle_update_channels_event(
    event: EventMessage, runner: Runner, state: PlayerState, **kwargs
):
    """event.msg == `PlayerState.get_raw_channels()`"""

    state.update_channels(event.msg)

    hls_channels_event(runner, state.get_raw_channels(), src=event.msg_src)


def handle_reset_sxm_event(
    event: EventMessage, runner: Runner, state: PlayerState, **kwargs
):
    """event.msg == None"""

    sxm_worker = runner.workers.get(ServerWorker.NAME)
    if sxm_worker is not None:
        sxm_worker.terminate()
        state.update_channels(None)
        cooldown = state.increase_cooldown()

        runner.log.warning(
            "SXM Client acting up, restarting it (cooldown: " f"{cooldown} seconds)"
        )

        del runner.workers[ServerWorker.NAME]

        state.sxm_running = False
        sxm_status_event(runner, Event.SXM_STATUS, state.sxm_running)


def handle_trigger_hls_stream_event(
    event: EventMessage,
    runner: Runner,
    state: PlayerState,
    host: str,
    port: int,
    output_folder: str,
    **kwargs,
):
    """event.msg == (channel_name: str, stream_protocol: str)"""

    hls_worker = runner.workers.get(HLSWorker.NAME)
    stream_folder: Optional[str] = None
    if output_folder is not None:
        stream_folder = os.path.join(output_folder, "streams")

    if hls_worker is not None:
        src_worker = runner.workers.get(event.msg_src)
        if src_worker is not None and src_worker.hls_stream_queue is not None:
            push_event(
                runner,
                src_worker,
                "hls_stream_queue",
                EventMessage(
                    event.msg_src, Event.HLS_STREAM_STARTED, state.stream_data
                ),
            )
            runner.log.info(
                f"Could not start new {HLSWorker.NAME}, one is "
                "already running passing "
                f"{Event.HLS_STREAM_STARTED} instead"
            )
        else:
            runner.log.warning(
                f"Could not start new {HLSWorker.NAME}, one is "
                "already running and no request was not HLSPlayer"
            )
    elif state.get_channel(event.msg[0]) is not None:
        runner.create_worker(
            HLSWorker,
            HLSWorker.NAME,
            ip=host,
            port=port,
            channel_id=event.msg[0],
            stream_folder=stream_folder,
            stream_protocol=event.msg[1],
            sxm_status=state.sxm_running,
        )
        state.stream_channel = event.msg
    else:
        runner.log.warning(
            f"Could not start new {HLSWorker.NAME}, invalid " f"channel id: {event.msg}"
        )


def handle_kill_hls_stream_event(
    event: EventMessage, runner: Runner, state: PlayerState, **kwargs
):
    """event.msg == None"""

    state.update_stream_data((None, None))

    hls_kill_event(runner, src=event.msg_src)
    for worker_name in (
        HLSWorker.NAME,
        ArchiveWorker.NAME,
        ProcessorWorker.NAME,
    ):
        worker = runner.workers.get(HLSWorker.NAME)
        if worker is not None:
            worker.full_stop()

            runner.log.info(f"Terminated {worker_name} worker")

            del runner.workers[worker_name]


def handle_hls_stream_started_event(
    event: EventMessage,
    runner: Runner,
    state: PlayerState,
    output_folder: str,
    reset_songs: bool,
    **kwargs,
):
    """event.msg == (channel_name: str, stream_url: str)"""

    stream_folder: Optional[str] = None
    archive_folder: Optional[str] = None
    processed_folder: Optional[str] = None
    if output_folder is not None:
        stream_folder = os.path.join(output_folder, "streams")
        archive_folder = os.path.join(output_folder, "archive")
        processed_folder = os.path.join(output_folder, "processed")

    state.update_stream_data(event.msg)
    hls_start_event(runner, state.stream_data, src=event.msg_src)

    if output_folder is not None:
        runner.create_worker(
            ArchiveWorker,
            ArchiveWorker.NAME,
            stream_folder=stream_folder,
            archive_folder=archive_folder,
            stream_data=state.stream_data,
            channels=state.get_raw_channels(),
            raw_live_data=state.get_raw_live(),
        )

        runner.create_worker(
            ProcessorWorker,
            ProcessorWorker.NAME,
            processed_folder=processed_folder,
            archive_folder=archive_folder,
            reset_songs=reset_songs,
            stream_data=state.stream_data,
            channels=state.get_raw_channels(),
            raw_live_data=state.get_raw_live(),
        )


def handle_update_metadata_event(
    event: EventMessage, runner: Runner, state: PlayerState, **kwargs
):
    """event.msg == (state.get_raw_live())"""

    state.stream_channel = event.msg["moduleResponse"]["liveChannelData"]["channelId"]
    state.update_live(event.msg)
    hls_metadata_event(runner, state.get_raw_live(), src=event.msg_src)


def handle_hls_stderror_lines_event(
    event: EventMessage, runner: Runner, state: PlayerState, **kwargs
):
    """event.msg == lines: List[str]"""

    do_reset = False
    for line in event.msg:
        runner.log.debug(f"ffmpeg STDERR: {line}")

        if "503" in line:
            do_reset = True

    if do_reset:
        handle_reset_sxm_event(event, runner, state)


def handle_debug_start_player_event(
    event: EventMessage, runner: Runner, state: PlayerState, **kwargs
):
    """event.msg ==
    (player_name: str,
     channel_id: str,
     filename: str,
     stream_protocol: str)"""

    player_name = event.msg[0]
    channel_id = event.msg[1]
    filename = event.msg[2]
    stream_protocol = event.msg[3]

    if state.stream_channel is not None and channel_id not in state.stream_channel:
        runner.log.warning(
            "Cannot start player, different HLS stream " f"playing: {state.stream_url}"
        )
    else:
        runner.create_worker(
            CLIPlayerWorker,
            player_name,
            filename=filename,
            stream_protocol=stream_protocol,
            sxm_status=state.sxm_running,
            stream_data=(channel_id, state.stream_url),
            channels=state.get_raw_channels(),
            raw_live_data=state.get_raw_live(),
        )


def handle_debug_stop_player_event(event: EventMessage, runner: Runner, **kwargs):
    """event.msg == player_name: str"""

    worker = runner.workers.get(event.msg)
    if worker is None:
        runner.log.warning(f"Debug Player {event.msg} is not currently running")
    else:
        worker.full_stop()
